{
  open Parser
  let shift_str s = String.sub s 1 ((String.length s) - 1)
  let eof_action = ref (fun () -> ())
  let set_eof_action f = eof_action := f
  let get_eof_action f = !eof_action
}

rule lexer = parse
| [' ' '\t' '\n' '\r'] { lexer lexbuf }
|"(*"([^ '*']*('*'[^ ')'])?)*("*)" | "*" | "") { lexer lexbuf }
| 'U'['0'-'9']* { UNI (int_of_string ("0" ^ shift_str (Lexing.lexeme lexbuf))) }
| 'L'['0'-'9']* { LIN (int_of_string ("0" ^ shift_str (Lexing.lexeme lexbuf))) }
| '#'['0'-'9']* { USHIFT (int_of_string (let s = shift_str (Lexing.lexeme lexbuf) in
                                          if s = "" then "1" else s)) }
| "let" { LET }
| ";;"
| ";" { EOC }
| '.' { DOT }
| ['P' 'p']"rint" { PRINT }
| ['C' 'c']"ompute" { COMPUTE }
| ['C' 'c']"heck" { CHECK }
| ['A' 'a']"xiom" { AXIOM }
| ['I' 'i']"nclude" { INCLUDE }
| ['I' 'i']"mport" { IMPORT }
| ['I' 'i']"nstantiate" { INSTANTIATE }
| ['G' 'g']"eneralize" { GENERALIZE }
| ['R' 'r']"es"("tart"|"et") { RESTART }
| ['D' 'd']"ump" { DUMP }
| ['R' 'r']"aw" { RAW }
| ['I' 'i']"nterpret" { INTERPRET }
| ['F' 'f']"orget" { FORGET }
| "val" { DEF }
| "Eq" { EQ }
| ":=" { EQDEF }
| '=' { EQSIGN }
| '+' { PLUS }
| "inl" { INL }
| "inr" { INR }
| "Ind"("Plus"|'+') | "PlusInd" { PLUSIND }
| "UnitInd" | "IndUnit" { UNITIND }
| "in" { IN }
| "LPi"| "LΠ" { LPI }
| "Pi" | "Π" { PI }
| "Sig" |"Σ" { SIG }
| "fun" | "λ" { LAM }
| "->" | "⟶" { ARR }
| "×" { TIMES }
| "|->" | "⟼" { POINTSTO }
| "Loc" { LOC }
| "new" { NEW }
| "get" { GET }
| "set" { SET }
| "free" { FREE }
| "value" { VAL }
| "T" { TTY }
| "fst" | "π₁" { FST }
| "snd" | "π₂" { SND }
| "refl" { REFL }
| ['P' 'p']"ack" { PACK }
| "Unit" { UNITTY }
| '<' { LABRA }
| '>' { RABRA }
| '[' { LSQBRA }
| ']' { RSQBRA }
| "Forall" | "∀" { FORALL }
| "Exists" | "∃" { EXISTS }
| ['A' 'a']"ny" { ANY }
| '"'(_#'"')*'"' { STR ((fun s -> String.sub s 1 ((String.length s) - 2)) (Lexing.lexeme lexbuf)) }
| '(' { LPAR }
| ')' { RPAR }
| '!' { BANG }
| '[' { LBRA }
| ']' { RBRA }
| "-o" | "⊸" { LOLLI }
| '*' | "⊗" { TENSOR }
| '&' { WITH }
| ',' { COMMA }
| "::" { DCOL }
| ':' { COL }
| "Fty" { FTY }
| 'F' { FPAIR }
| "WInd"
| "IndW" { WIND }
| 'W' { W }
| "sup" { SUP }
| 'I' { ITY }
| 'J' { JELIM }
| "NatInd"
| "IndNat"
| "Indℕ" { NATIND }
| "Nat" | "ℕ" { NAT }
| "Bot" | "⊥" { BOT }
| "Top" | "⊤" { TOP }
| "IndBot" | "Ind⊥" 
| "BotInd" { BOTIND }
| "S" { SUCC }
| ['0'-'9']+ { INT (int_of_string (Lexing.lexeme lexbuf)) }
| "Comp" { GTY }
| 'G' { G }
| "run" { RUNG }
| '_' { DCVAR (Lexing.lexeme lexbuf) }
| ['a'-'z' 'A'-'Z']['a'-'z' 'A'-'Z' '_' '\'' '0'-'9']* { VAR (Lexing.lexeme lexbuf)}
| eof { (!eof_action)(); EOF }
