%{
  open Prelude
  open Ast
  let annot t (TIn t') = TIn (Annot (t, t'))
  let rec bnd x e = match x with
                           [] -> e
                         | x :: t -> let TIn r = bnd t e in
                                     TIn (Bnd (x, r))
  let var x = TIn (Var x)
  let app t t' = TIn (App (t,t'))
  let lam x t = TIn (Lam (bnd [x] t))
  let fpair t t' = TIn (F (t, t'))
  let pair t t' = TIn (Pair (t, t'))
  let wpair t t' = TIn (WPair (t, t'))

  let letpair a b t t' = TIn (LetPair (t, bnd [a; b] t'))
  let letunit t t' = TIn (LetUnit (t, t'))
  let letf x a t t' = TIn (LetF (t, bnd [x; a] t'))
  let letsyn x t t' = TIn (Let (t, bnd [x] t'))
  let letirrunit t t' = TIn (LetIrrUnit (t, t'))
  let letirrpair a b t t' = TIn (LetIrrPair (t, bnd [a; b] t'))
  let letpack a b t t' = TIn (LetPack (t, bnd [a; b] t'))
  let letirr a t t' = TIn (LetIrr (t, bnd [a] t'))
  let fst' t = TIn (Fst t)
  let snd' t = TIn (Snd t)
  let g t = TIn (G t)
  let rung t = TIn (RunG t)

  let pi x t t' = TIn (Pi (t, bnd [x] t'))
  let sigma x t t' = TIn (Sigma (t, bnd [x] t'))
  let fty x t t' = TIn (Fty (t, bnd [x] t'))
  let wty x t t' = TIn (W (t, bnd [x] t'))
  let lpi x t t' = TIn (LPi (t, bnd [x] t'))
  let forall x t t' = TIn (Forall (t, bnd [x] t'))
  let exists x t t' = TIn (Exists (t, bnd [x] t'))
  let rec promq q l t t' = match l with
                            [] -> bnd [] t'
                           | h :: tl -> let t' = promq q tl t t' in
                                        q h t t'
  let promq q (x,l) t t' = promq q (x :: l) t t'

  let je x y p ty t e = TIn (J (bnd [x; y; p] ty, t, e))
  let eq ty x y = TIn (Eq (ty, x, y))

  let tensor t t' = TIn (Tensor (t, t'))
  let lolli t t' = TIn (Lolli (t, t'))
  let withty t t' = TIn (With (t, t'))
  let uni i = TIn (Uni i)
  let lin i = TIn (Lin i)

  let zero = TIn (Zero)
  let succ x = TIn (Succ x)
  let natind x ty t y z t' n = TIn (Natind (bnd [x] ty, t, bnd [y; z] t', n))

  let sup a b = TIn (Sup (a, b))
  let wind x ty a b r t w = TIn (Wind (bnd [x] ty, bnd [a; b; r] t, w))

  let inl x = TIn (Inl x)
  let inr x = TIn (Inr x)
  let plus a b = TIn (Plus (a, b))
  let plusind x ty a ca b cb w = TIn (Plusind (bnd [x] ty, bnd [a] ca, bnd [b] cb, w))

  let unitind x ty t w = TIn (Unitind (bnd [x] ty, t, w))

  let rec nat_of_int = function 0 -> TIn Zero | n -> TIn (Succ (nat_of_int (n-1)))

  let irr a = TIn (Irr a)
  let squash a = TIn (Squash a)
  let pack a b = TIn (Pack (a, b))

  let letget a b t t'' =  match t with
                           TIn (Get (t, t')) -> TIn (LetGet (t, t', bnd [a; b] t''))
                          | _ -> raise Parsing.Parse_error
  let set t t' t'' = TIn (Assign (t', t, t''))
  let letval a t t' = TIn (LetT (t, bnd [a] t'))

  let pointsto l x = TIn (Pointsto (l, x))


  type command = Axiom of var * tm
               | Def of var * tm * tm
               | Definfer of var * tm
               | Infer of tm
               | Compute of tm
               | Include of string
               | Import of string
               | Dump of string
               | Print
               | Restart
               | Instantiate of var * tm
               | Generalize of var
               | Raw of tm
               | Interpret of tm
               | Forget of var

  let gensym =
   let int_of_letter c = (int_of_char c.[0]) - (int_of_char 'a') in
   let cnt = Array.make 26 0 in
   let string_of_int = function 0 -> "" | n -> string_of_int (n-1) in
   fun () -> let a = random_letter () in
             let c = int_of_letter a in
             let s = "%" ^ a ^ (string_of_int (cnt.(c))) in
             cnt.(c) <- cnt.(c) + 1;
             s
  type lptrans = LPVar of string | LPTrans of (Ast.tm -> Ast.tm -> Ast.tm)
  let lpstr = function LPVar s -> s | _ -> gensym ()
  let lptr s = function LPVar _ -> id | LPTrans f -> fun g x y -> g x (f (TIn (Var s)) y)
%}

%token ANY
%token POINTSTO VAL GET SET FREE NEW TTY LOC BANG
%token LABRA RABRA LSQBRA RSQBRA FORALL EXISTS PACK
%token AXIOM EOC DEF CHECK COMPUTE EQDEF EQ PRINT INCLUDE RESTART
%token IMPORT INSTANTIATE GENERALIZE RAW DUMP INTERPRET FORGET
%token LPAR RPAR LBRA RBRA COMMA COL DOT DCOL
%token ARR LET EQSIGN IN FST SND LAM G W
%token PLUS PLUSIND INL INR UNITIND TIMES
%token JELIM REFL NATIND NAT SUCC BOT BOTIND WIND
%token <int> INT
%token PI SIG UNITTY LOLLI TENSOR WITH FTY LPI FPAIR ITY GTY RUNG SUP TOP
%token <int> UNI LIN USHIFT
%token <string> VAR STR DCVAR
%token EOF
%token ILLEGAL


%left COL DCOL
%left PLUS
%left TENSOR
%left WITH
%right LOLLI
%left TIMES
%left FST SND VAR
%right RUNG
%nonassoc LPAR RPAR LBRA RBRA COMMA LET EQSIGN IN
%nonassoc LAM FPAIR ITY FTY
%nonassoc AXIOM EOC DEF CHECK COMPUTE EQDEF
%nonassoc JELIM EQ REFL
%right ARR

%start e
%start comm
%start file
%type <Ast.tm> e
%type <command> comm
%type <Ast.var * Ast.var list> list_var
%type <command list> file

%%

file :
  | EOF { [] }
  | stmt file { $1 :: $2 }

stmt :
  | DEF VAR COL e EQDEF e   { Def ($2, $6, $4) }
  | DEF VAR EQDEF e         { Definfer ($2, $4) }
  | AXIOM VAR COL e         { Axiom ($2, $4) }
  | CHECK e                 { Infer $2 }
/*| CHECK e COL e           { Check ($2, $4) } */
  | FORGET VAR              { Forget $2 }
  | COMPUTE e               { Compute $2 }
  | INCLUDE STR             { Include $2 }
  | IMPORT STR              { Import  $2 }
  | IMPORT VAR              { Import  $2 }
  | PRINT                    { Print }
  | RESTART                 { Restart }
  | INSTANTIATE VAR EQDEF e { Instantiate ($2, $4) }
  | GENERALIZE VAR          { Generalize $2 }
  | RAW e                   { Raw $2 }
  | DUMP STR                { Dump $2 }
  | INTERPRET e             { Interpret $2 }
comm :
  | stmt EOC { $1 }

e :
  | t COL e             { annot $1 $3 } 
/*  | q list_var COL e DOT e { promq $1 $2 $4 $6 } */
  | q list_q DOT e      { $2 $1 $4 }
  | a b e                { $2 $1 $3}
  | LAM list_nlp DOT e   { let elam = function
                                          LPVar s -> lam s
                                         |LPTrans f -> let s = gensym () in
                                                       fun y -> lam s (f (TIn (Var s)) y) in 
                           elam (fst $2) (List.fold_right elam (snd $2) $4) }
  | LAM list_nlp ARR e   { let elam = function
                                          LPVar s -> lam s
                                         |LPTrans f -> let s = gensym () in
                                                       fun y -> lam s (f (TIn (Var s)) y) in 
                           elam (fst $2) (List.fold_right elam (snd $2) $4) }
/*  | LET lp EQSIGN e IN e { $2 $4 $6} */
  | LET nlp EQSIGN e IN e { (match $2 with
                              LPVar x -> letsyn x
                             |LPTrans f -> f) $4 $6 }
  | GET LPAR e COMMA e RPAR { TIn (Get ($3, $5)) }
/*| LET LPAR BANG VAR COMMA VAR RPAR EQSIGN GET LPAR e COMMA e RPAR IN e
                         { letget $4 $6 $11 $13 $16 } */
  | FREE LPAR e COMMA e RPAR { TIn (Free ($3, $5)) }
  | NEW e                { TIn (New $2) }
  | FST e                { fst' $2}
  | SND e                { snd' $2 }
  | USHIFT e             { TIn (UShift ($1, $2)) }
  | INL e                { inl $2 }
  | INR e                { inr $2 }
  | GTY e                { TIn (Gty $2) }
  | G   e                { g $2 }
  | TTY e                { TIn (T $2) }
  | VAL e                { TIn (Val $2) }
  | SET LPAR e COMMA e COMMA e RPAR
                         { set $3 $5 $7 }
  | SUCC e               { succ $2 }
  | JELIM LPAR bv DOT bv DOT bv DOT e COMMA e COMMA e RPAR
                         { je $3 $5 $7 $9 $11 $13 }
  | NATIND LPAR bv DOT e COMMA e COMMA bv DOT bv DOT e COMMA e RPAR
                         { natind $3 $5 $7 $9 $11 $13 $15 }
  | NATIND LPAR e COMMA e COMMA bv DOT bv DOT e COMMA e RPAR
                         { natind "_" $3 $5 $7 $9 $11 $13 }
  | WIND LPAR bv DOT e COMMA bv DOT bv DOT bv DOT e COMMA e RPAR
                         { wind $3 $5 $7 $9 $11 $13 $15 }
  | WIND LPAR e COMMA bv DOT bv DOT bv DOT e COMMA e RPAR
                         { wind "_" $3 $5 $7 $9 $11 $13 }
  | BOTIND e             { TIn (Botind $2) }
  | PLUSIND LPAR bv DOT e COMMA bv DOT e COMMA bv DOT e COMMA e RPAR
                         { plusind $3 $5 $7 $9 $11 $13 $15 }
  | PLUSIND LPAR e COMMA bv DOT e COMMA bv DOT e COMMA e RPAR
                         { plusind "_" $3 $5 $7 $9 $11 $13 }
  | UNITIND LPAR bv DOT e COMMA e COMMA e RPAR
                         { unitind $3 $5 $7 $9 }
  | UNITIND LPAR e COMMA e COMMA e RPAR
                         { unitind "_" $3 $5 $7 }
  | RUNG e               { rung $2 }
  | a                    { $1 }
  | t DCOL e             { annot $1 $3 }

a :
  | t {
    $1 }
  | a t {
    app $1 $2 }
  | EQ t t t             { eq $2 $3 $4 }
t :
  | VAR                       { var $1}
  | LPAR e RPAR               { $2 }
  | LSQBRA e RSQBRA      { irr $2 }
  | LPAR e COMMA e RPAR       { pair $2 $4 }
  | LSQBRA e COMMA e RSQBRA       { wpair $2 $4 }
  | LABRA e RABRA             { squash $2 }
  | FPAIR LPAR e COMMA e RPAR { fpair $3 $5 }
  | SUP LPAR e COMMA e RPAR   { sup $3 $5 }
  | PACK LPAR e COMMA e RPAR  { pack $3 $5 }
  | LPAR RPAR                 { TIn Unit }
  | UNITTY                    { TIn Unitty }
  | ITY                       { TIn I }
  | INT                       { nat_of_int $1 }
  | UNI                       { uni $1 }
  | LIN                       { lin $1 }
  | REFL                      { TIn Refl }
  | NAT                       { TIn Nat }
  | BOT                       { TIn Bot }
  | TOP                       { TIn Top }
  | LOC                       { TIn Loc }
  | ANY                       { TIn Any }

/* let pattern
lp :
  | LPAR VAR COMMA VAR RPAR       { letpair $2 $4}
  | FPAIR LPAR VAR COMMA VAR RPAR { letf $3 $5}
  | LPAR RPAR                     { letunit }
  | VAR                           { letsyn $1 } 
  | LABRA LPAR RPAR RABRA         { letirrunit }
  | LABRA VAR COMMA VAR RABRA     { letirrpair $2 $4 }
  | LABRA VAR RABRA               { letirr $2}
  | PACK LPAR VAR COMMA VAR RPAR  { letpack $3 $5 }
  | VAL VAR                       { letval $2 }
*/

/* nested let pattern */
nlp :
  | LPAR nlp COMMA nlp RPAR       { let s = lpstr $2 in
                                    let s' = lpstr $4 in
                                    LPTrans (lptr s' $4 (lptr s $2 (letpair s s'))) }
  | FPAIR LPAR nlp COMMA nlp RPAR { let s = lpstr $3 in
                                    let s' = lpstr $5 in
                                    LPTrans (lptr s' $5 (lptr s $3 (letf s s')))}
  | LPAR RPAR                     { LPTrans letunit }
  | bv                           { LPVar $1 } 
  | LABRA RABRA         { LPTrans letirrunit }
  | LABRA nlp COMMA nlp RABRA     { let s  = lpstr $2 in
                                    let s' = lpstr $4 in
                                    LPTrans (lptr s' $4 (lptr s $2 (letirrpair s s'))) }
  | LABRA nlp RABRA               { let a = $2 in
                                    let s = lpstr a in
                                    LPTrans (lptr s a (letirr s))}
  | PACK LPAR nlp COMMA nlp RPAR  { let s  = lpstr $3 in
                                    let s' = lpstr $5 in
                                    LPTrans (lptr s $3 (lptr s' $5 (letpack s s'))) }
  | VAL nlp                       { let s = lpstr $2 in
                                    LPTrans (lptr s $2 (letval s)) }
  | LPAR BANG nlp COMMA nlp RPAR  { let s = lpstr $3 in
                                    let s' = lpstr $5 in
                                    LPTrans (lptr s $3 (lptr s' $5 (letget s s'))) }
list_var :
  | bv list_var { ($1, (fun (a,b) -> (a :: b)) $2) }
  | bv          { ($1, []) }

list_q :
  | list_var COL e { fun q -> promq q $1 $3 }
  | list_var COL e COMMA list_q { fun q e -> promq q $1 $3 ($5 q e) }

list_nlp :
  | nlp list_nlp { ($1, (fun (a,b) -> (a :: b)) $2) }
  | nlp          { ($1, []) }

/* binary type connectives */
b :
  | LOLLI    { lolli }
  | TENSOR   { tensor }
  | WITH     { withty }
  | ARR      { pi "_" }
  | PLUS     { plus }
  | POINTSTO { pointsto }
  | TIMES    { sigma "_" }

/* quantifiers */
q :
  | PI { pi }
  | SIG { sigma }
  | FTY { fty }
  | W { wty }
  | LPI { lpi }
  | FORALL { forall }
  | EXISTS { exists }

/* binder variables */
bv :
  | VAR { $1 }
  | DCVAR { $1 }

%%
